import pandas as pd

from kabutobashi.domain.errors import KabutobashiBlockSeriesIsNoneError

from ..decorator import block
from .abc_parameterize_block import get_impact


@block(
    block_name="parameterize_sma",
    series_required_columns=["sma_short", "sma_medium", "sma_long", "close", "sma_buy_signal", "sma_sell_signal"],
)
class ParameterizeSmaBlock:
    series: pd.DataFrame
    influence: int = 2
    tail: int = 5

    def _process(self) -> dict:
        df = self.series
        if df is None:
            raise KabutobashiBlockSeriesIsNoneError()
        df["sma_short_diff"] = (df["sma_short"] - df["close"]) / df["sma_short"]
        df["sma_medium_diff"] = (df["sma_medium"] - df["close"]) / df["sma_medium"]
        df["sma_long_diff"] = (df["sma_long"] - df["close"]) / df["sma_long"]
        # difference from sma_long
        df["sma_long_short"] = (df["sma_long"] - df["sma_short"]) / df["sma_long"]
        df["sma_long_medium"] = (df["sma_long"] - df["sma_medium"]) / df["sma_long"]
        params = {
            "sma_short_diff": df["sma_short_diff"].tail(3).mean(),
            "sma_medium_diff": df["sma_medium_diff"].tail(3).mean(),
            "sma_long_diff": df["sma_long_diff"].tail(3).mean(),
            "sma_long_short": df["sma_long_short"].tail(3).mean(),
            "sma_long_medium": df["sma_long_medium"].tail(3).mean(),
            "sma_impact": get_impact(df=df, influence=self.influence, tail=self.tail, prefix="sma"),
            "dt": max(df.index.to_list()),
        }

        return params



import httpx

from kabutobashi.domain.errors import KabutobashiPageError
from kabutobashi.domain.values import UserAgent

__all__ = ["from_url"]


def from_url(url: str) -> str:
    user_agent = UserAgent.get_user_agent_header()
    r = httpx.get(url, headers=user_agent)

    if r.status_code != 200:
        raise KabutobashiPageError(url=url)

    return r.text



from .crawl_stock_info_block import CrawlStockInfoBlock
from .crawl_stock_info_multiple_days_block import CrawlStockInfoMultipleDays2Block, CrawlStockInfoMultipleDaysBlock
from .crawl_stock_ipo_block import CrawlStockIpoBlock



import pandas as pd

from kabutobashi.domain.errors import KabutobashiBlockSeriesIsNoneError

from ..decorator import block
from .abc_parameterize_block import get_impact


@block(block_name="parameterize_momentum", series_required_columns=["momentum_buy_signal", "momentum_sell_signal"])
class ParameterizeMomentumBlock:
    series: pd.DataFrame
    influence: int = 2
    tail: int = 5

    def _process(self) -> dict:
        df = self.series
        if df is None:
            raise KabutobashiBlockSeriesIsNoneError()
        params = {
            "momentum_impact": get_impact(df=df, influence=self.influence, tail=self.tail, prefix="momentum"),
            "dt": max(df.index.to_list()),
        }

        return params



import pandas as pd

from ..decorator import block

__all__ = ["ProcessBollingerBandsBlock"]


@block(block_name="process_bollinger_bands", series_required_columns=["close"])
class ProcessBollingerBandsBlock:
    series: pd.DataFrame
    band_term: int = 12
    continuity_term: int = 10

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(mean=df["close"].rolling(self.band_term).mean(), std=df["close"].rolling(self.band_term).std())
        df = df.assign(
            upper_1_sigma=df.apply(lambda x: x["mean"] + x["std"] * 1, axis=1),
            lower_1_sigma=df.apply(lambda x: x["mean"] - x["std"] * 1, axis=1),
            upper_2_sigma=df.apply(lambda x: x["mean"] + x["std"] * 2, axis=1),
            lower_2_sigma=df.apply(lambda x: x["mean"] - x["std"] * 2, axis=1),
            upper_3_sigma=df.apply(lambda x: x["mean"] + x["std"] * 3, axis=1),
            lower_3_sigma=df.apply(lambda x: x["mean"] - x["std"] * 3, axis=1),
        )
        return df

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(
            over_upper=df.apply(lambda x: 1 if x["close"] > x["upper_2_sigma"] else 0, axis=1),
            over_lower=df.apply(lambda x: 1 if x["close"] < x["lower_2_sigma"] else 0, axis=1),
            over_upper_continuity=lambda x: x["over_upper"].rolling(self.continuity_term).sum(),
            over_lower_continuity=lambda x: x["over_lower"].rolling(self.continuity_term).sum(),
        )

        df["bollinger_bands_buy_signal"] = df["over_upper"].apply(lambda x: 1 if x > 0 else 0)
        df["bollinger_bands_sell_signal"] = df["over_lower"].apply(lambda x: 1 if x > 0 else 0)
        return df

    def _process(self) -> pd.DataFrame:

        applied_df = self._apply(df=self.series)
        signal_df = self._signal(df=applied_df)
        required_columns = [
            "upper_1_sigma",
            "lower_1_sigma",
            "upper_2_sigma",
            "lower_2_sigma",
            "upper_3_sigma",
            "lower_3_sigma",
            "over_upper_continuity",
            "over_lower_continuity",
            "bollinger_bands_buy_signal",
            "bollinger_bands_sell_signal",
        ]
        return signal_df[required_columns]



import pandas as pd

from ..decorator import block


@block(block_name="process_psycho_logical", series_required_columns=["close"])
class ProcessPsychoLogicalBlock:
    series: pd.DataFrame
    psycho_term: int = 12
    upper_threshold: float = 0.75
    lower_threshold: float = 0.25

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        df_ = df.copy()
        df_["shift_close"] = df_["close"].shift(1)
        df_ = df_.fillna(0)
        df_["diff"] = df_.apply(lambda x: x["close"] - x["shift_close"], axis=1)

        df_["is_raise"] = df_["diff"].apply(lambda x: 1 if x > 0 else 0)

        df_["psycho_sum"] = df_["is_raise"].rolling(self.psycho_term).sum()
        df_["psycho_line"] = df_["psycho_sum"].apply(lambda x: x / self.psycho_term)

        df_["bought_too_much"] = df_["psycho_line"].apply(lambda x: 1 if x > self.upper_threshold else 0)
        df_["sold_too_much"] = df_["psycho_line"].apply(lambda x: 1 if x < self.lower_threshold else 0)
        return df_

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        df["psycho_logical_buy_signal"] = df["sold_too_much"]
        df["psycho_logical_sell_signal"] = df["bought_too_much"]
        return df

    def _process(self) -> pd.DataFrame:

        applied_df = self._apply(df=self.series)
        signal_df = self._signal(df=applied_df)
        required_columns = [
            "psycho_line",
            "bought_too_much",
            "sold_too_much",
            "psycho_logical_buy_signal",
            "psycho_logical_sell_signal",
        ]
        return signal_df[required_columns]



from dataclasses import dataclass, field, replace
from typing import List, Union

from kabutobashi.domain.entity.blocks.basis_blocks import BlockGlue, BlockOutput, IBlock
from kabutobashi.domain.entity.blocks.decorator import block_from


@dataclass(frozen=True)
class Flow:
    block_glue: BlockGlue

    @staticmethod
    def from_json(params_list: List[dict]) -> "Flow":
        flow_params = {}
        block_list = []
        for params in params_list:
            block = block_from(params["block_name"])
            block_list.append(block)
            flow_params.update({params["block_name"]: params.get("params", {})})
        return Flow.initialize(params=flow_params).then(block=block_list)

    @staticmethod
    def initialize(params: dict) -> "Flow":
        initial_output = BlockOutput(series=None, params=params, block_name="initial_output", execution_order=1)
        glue = BlockGlue(series=None, params=params, block_outputs={"FLOW_INITIAL": initial_output})
        return Flow(block_glue=glue)

    def then(self, block: Union[type[IBlock], List[type[IBlock]]]) -> "Flow":
        if type(block) is list:
            flow = self
            glue: BlockGlue = self.block_glue
            for b in block:
                glue = b.glue(glue=glue)
                flow = replace(flow, block_glue=glue)
            return flow
        else:
            new_glue = block.glue(glue=self.block_glue)
            return replace(self, block_glue=new_glue)

    def reduce(self, block: IBlock) -> BlockGlue:
        new_glue = block.glue(glue=self.block_glue)
        return new_glue


@dataclass(frozen=True)
class FlowPath:
    next_sequence_no: int = field(default=1)
    flow_params_list: list = field(default_factory=list)

    def sma(self) -> "FlowPath":
        next_sequence_no = self.next_sequence_no
        flow_params_list = self.flow_params_list
        process = {"id": "process_sma", "block_name": "process_sma", "sequence_no": next_sequence_no, "params": {}}
        parameterize = {
            "id": "parameterize_sma",
            "block_name": "parameterize_sma",
            "sequence_no": next_sequence_no + 1,
            "params": {},
        }
        flow_params_list.extend([process, parameterize])
        return replace(self, next_sequence_no=next_sequence_no + 2, flow_params_list=flow_params_list)

    def macd(self) -> "FlowPath":
        next_sequence_no = self.next_sequence_no
        flow_params_list = self.flow_params_list

        process = {"id": "process_macd", "block_name": "process_macd", "sequence_no": next_sequence_no, "params": {}}
        parameterize = {
            "id": "parameterize_macd",
            "block_name": "parameterize_macd",
            "sequence_no": next_sequence_no + 1,
            "params": {},
        }
        flow_params_list.extend([process, parameterize])
        return replace(self, next_sequence_no=next_sequence_no + 2, flow_params_list=flow_params_list)

    def stochastics(self) -> "FlowPath":
        next_sequence_no = self.next_sequence_no
        flow_params_list = self.flow_params_list

        process = {
            "id": "process_stochastics",
            "block_name": "process_stochastics",
            "sequence_no": next_sequence_no,
            "params": {},
        }
        parameterize = {
            "id": "parameterize_stochastics",
            "block_name": "parameterize_stochastics",
            "sequence_no": next_sequence_no + 1,
            "params": {},
        }
        flow_params_list.extend([process, parameterize])
        return replace(self, next_sequence_no=next_sequence_no + 2, flow_params_list=flow_params_list)

    def momentum(self) -> "FlowPath":
        next_sequence_no = self.next_sequence_no
        flow_params_list = self.flow_params_list

        process = {
            "id": "process_momentum",
            "block_name": "process_momentum",
            "sequence_no": next_sequence_no,
            "params": {},
        }
        parameterize = {
            "id": "parameterize_momentum",
            "block_name": "parameterize_momentum",
            "sequence_no": next_sequence_no + 1,
            "params": {},
        }
        flow_params_list.extend([process, parameterize])
        return replace(self, next_sequence_no=next_sequence_no + 2, flow_params_list=flow_params_list)

    def adx(self) -> "FlowPath":
        next_sequence_no = self.next_sequence_no
        flow_params_list = self.flow_params_list

        process = {"id": "process_adx", "block_name": "process_adx", "sequence_no": next_sequence_no, "params": {}}
        parameterize = {
            "id": "parameterize_adx",
            "block_name": "parameterize_adx",
            "sequence_no": next_sequence_no + 1,
            "params": {},
        }
        flow_params_list.extend([process, parameterize])
        return replace(self, next_sequence_no=next_sequence_no + 2, flow_params_list=flow_params_list)

    def psycho_logical(self) -> "FlowPath":
        next_sequence_no = self.next_sequence_no
        flow_params_list = self.flow_params_list

        process = {
            "id": "process_psycho_logical",
            "block_name": "process_psycho_logical",
            "sequence_no": next_sequence_no,
            "params": {},
        }
        parameterize = {
            "id": "parameterize_psycho_logical",
            "block_name": "parameterize_psycho_logical",
            "sequence_no": next_sequence_no + 1,
            "params": {},
        }
        flow_params_list.extend([process, parameterize])
        return replace(self, next_sequence_no=next_sequence_no + 2, flow_params_list=flow_params_list)

    def bollinger_bands(self) -> "FlowPath":
        next_sequence_no = self.next_sequence_no
        flow_params_list = self.flow_params_list

        process = {
            "id": "process_bollinger_bands",
            "block_name": "process_bollinger_bands",
            "sequence_no": next_sequence_no,
            "params": {},
        }
        parameterize = {
            "id": "parameterize_bollinger_bands",
            "block_name": "parameterize_bollinger_bands",
            "sequence_no": next_sequence_no + 1,
            "params": {},
        }
        flow_params_list.extend([process, parameterize])
        return replace(self, next_sequence_no=next_sequence_no + 2, flow_params_list=flow_params_list)

    def volatility(self) -> "FlowPath":
        next_sequence_no = self.next_sequence_no
        flow_params_list = self.flow_params_list

        parameterize = {
            "id": "parameterize_volatility",
            "block_name": "parameterize_volatility",
            "sequence_no": next_sequence_no + 1,
            "params": {},
        }
        flow_params_list.extend([parameterize])
        return replace(self, next_sequence_no=next_sequence_no + 1, flow_params_list=flow_params_list)

    def pct_change(self) -> "FlowPath":
        next_sequence_no = self.next_sequence_no
        flow_params_list = self.flow_params_list

        parameterize = {
            "id": "parameterize_pct_change",
            "block_name": "parameterize_pct_change",
            "sequence_no": next_sequence_no + 1,
            "params": {},
        }
        flow_params_list.extend([parameterize])
        return replace(self, next_sequence_no=next_sequence_no + 1, flow_params_list=flow_params_list)

    def read_example(self, code: int) -> "FlowPath":
        next_sequence_no = self.next_sequence_no
        flow_params_list = self.flow_params_list
        flow_params_list.append(
            {
                "id": "read_example",
                "block_name": "read_example",
                "sequence_no": next_sequence_no,
                "params": {"code": code},
            }
        )
        return replace(self, next_sequence_no=next_sequence_no + 1, flow_params_list=flow_params_list)

    def read_sqlite3(self, code: int, database_dir) -> "FlowPath":
        next_sequence_no = self.next_sequence_no
        flow_params_list = self.flow_params_list
        flow_params_list.append(
            {
                "id": "read_sqlite3",
                "block_name": "read_sqlite3",
                "sequence_no": next_sequence_no,
                "params": {"code": code, "database_dir": database_dir},
            }
        )
        return replace(self, next_sequence_no=next_sequence_no + 1, flow_params_list=flow_params_list)

    def apply_default_pre_process(self) -> "FlowPath":
        next_sequence_no = self.next_sequence_no
        flow_params_list = self.flow_params_list
        flow_params_list.append(
            {"id": "default_pre_process", "block_name": "default_pre_process", "sequence_no": next_sequence_no}
        )
        return replace(self, next_sequence_no=next_sequence_no + 1, flow_params_list=flow_params_list)

    def dumps(self) -> List[dict]:
        return self.flow_params_list

    def execute(self) -> Flow:
        return Flow.from_json(params_list=self.flow_params_list)



from datetime import datetime
from typing import Tuple

import pandas as pd
from bs4 import BeautifulSoup

from ..decorator import block


@block(block_name="extract_stock_info_multiple_days", params_required_keys=["html_text"])
class ExtractStockInfoMultipleDaysBlock:
    main_html_text: str
    sub_html_text: str
    code: str

    def _decode(self, code: str, main_html_text: str, sub_html_text: str) -> dict:
        result_1 = []
        result_2 = []
        main_soup = BeautifulSoup(main_html_text, features="lxml")
        sub_soup = BeautifulSoup(sub_html_text, features="lxml")
        stock_recordset_tag = "md_card md_box"

        # ページの情報を取得
        stock_recordset = main_soup.find("div", {"class": stock_recordset_tag})
        mapping = {0: "dt", 1: "open", 2: "high", 3: "low", 4: "close", 5: "調整後終値", 6: "volume"}
        for tr in stock_recordset.find_all("tr"):
            tmp = {}
            for idx, td in enumerate(tr.find_all("td")):
                tmp.update({mapping[idx]: td.get_text()})
            result_1.append(tmp)

        # そのほかの情報
        stock_board = sub_soup.find("div", {"class": "md_card md_box mzp"})
        mapping2 = {0: "dt", 1: "psr", 2: "per", 3: "pbr", 4: "配当利回り(%)", 5: "close", 6: "調整後終値", 7: "volume"}
        for tr in stock_board.find_all("tr"):
            tmp = {}
            for idx, td in enumerate(tr.find_all("td")):
                tmp.update({mapping2[idx]: td.get_text()})
            result_2.append(tmp)

        df1 = pd.DataFrame(result_1).dropna()
        df2 = pd.DataFrame(result_2).dropna()

        df1 = df1[["dt", "open", "high", "low", "close"]]
        df2 = df2[["dt", "psr", "per", "pbr", "volume"]]

        df = pd.merge(df1, df2, on="dt")
        df["code"] = code
        return {"info_list": df.to_dict(orient="records")}

    def _process(self) -> Tuple[pd.DataFrame, dict]:
        result = self._decode(code=self.code, main_html_text=self.main_html_text, sub_html_text=self.sub_html_text)
        # to_df
        df = pd.DataFrame(data=result["info_list"])
        df.index = df["dt"]
        return df, result

    # def _validate_output(self, series: Optional[pd.DataFrame], params: Optional[dict]):
    #     keys = series.keys()
    # assert "info_list" in keys, "StockInfoMultipleDaysExtractBlockOutput must have 'info_list' column"


@block(block_name="extract_stock_info_multiple_days_2", params_required_keys=["main_html_text", "code"])
class ExtractStockInfoMultipleDays2Block:
    main_html_text: str
    code: str

    def _decode(self, code: str, main_html_text: str) -> dict:
        main_soup = BeautifulSoup(main_html_text, features="lxml")

        # 名前取得
        code_and_name = main_soup.find("div", {"class": "si_i1_1"}).find("h2")
        code_and_name.find("span").extract()
        name = code_and_name.get_text()
        # ページの情報を取得
        stock_recordset = main_soup.find("table", {"class": "stock_kabuka_dwm"})
        result_1 = []
        mapping = {0: "open", 1: "high", 2: "low", 3: "close", 4: "diff", 5: "ratio", 6: "volume"}
        for tr in stock_recordset.find_all("tr"):
            tmp = {}
            dt = tr.find("time")
            if dt:
                dt = datetime.strptime(dt.get_text(), "%y/%m/%d").strftime("%Y-%m-%d")
                tmp.update({"dt": dt})
            for idx, td in enumerate(tr.find_all("td")):
                tmp.update({mapping[idx]: td.get_text()})
            result_1.append(tmp)
        df = pd.DataFrame(result_1).dropna()
        df = df[["dt", "open", "high", "low", "close", "volume"]]

        df["code"] = code
        df["name"] = name
        return {"info_list": df.to_dict(orient="records")}

    def _process(self) -> Tuple[pd.DataFrame, dict]:
        result = self._decode(code=self.code, main_html_text=self.main_html_text)
        # to_df
        df = pd.DataFrame(data=result["info_list"])
        df.index = df["dt"]
        return df, result



from typing import List


class KabutobashiBaseError(Exception):
    pass


class KabutobashiPageError(KabutobashiBaseError):
    """
    Crawlが失敗したときに返すエラー
    """

    def __init__(self, url: str = ""):
        self.url = url

    def __str__(self):
        return f"error occurred when crawling [{self.url}]"


class KabutobashiEntityError(KabutobashiBaseError):
    pass


class KabutobashiBlockError(KabutobashiBaseError):
    pass


class KabutobashiBlockInstanceMismatchError(KabutobashiBlockError):
    pass


class KabutobashiBlockParamsIsNoneError(KabutobashiBlockError):
    pass


class KabutobashiBlockSeriesIsNoneError(KabutobashiBlockError):
    pass


class KabutobashiBlockSeriesDtIsMissingError(KabutobashiBlockError):
    def __init__(self, code: str, dt: List[str]):
        self.code = code
        self.dt = dt

    def __str__(self):
        return f"{self.dt} in {self.code} is missing"


class KabutobashiBlockGlueError(KabutobashiBlockError):
    """
    KabutobashiBlockGlueError is base error for `BlockGlue`.
    """

    pass


class KabutobashiBlockDecoratorError(KabutobashiBlockError):
    """
    KabutobashiBlockDecoratorError is base error for `@block` decorator.
    """

    pass


class KabutobashiBlockDecoratorNameError(KabutobashiBlockDecoratorError):
    """
    class must end with `Block`.
    """

    pass


class KabutobashiBlockDecoratorTypeError(KabutobashiBlockDecoratorError):
    """
    fist argument of @block() must be type
    """

    pass


class KabutobashiBlockDecoratorReturnError(KabutobashiBlockDecoratorError):
    """
    function-return-type is not matched.
    """

    pass


class KabutobashiBlockDecoratorNotImplementedError(KabutobashiBlockDecoratorError):
    """
    The function that was intended to be implemented has not been implemented.
    """

    pass



import pandas as pd

from ..decorator import block
from .abc_process_block import cross


@block(block_name="process_momentum", series_required_columns=["close"])
class ProcessMomentumBlock:
    series: pd.DataFrame
    term: int = 12

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:

        df = df.assign(
            momentum=df["close"].shift(10),
        ).fillna(0)
        df = df.assign(sma_momentum=lambda x: x["momentum"].rolling(self.term).mean())
        return df

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.join(cross(df["sma_momentum"]))
        df = df.rename(columns={"to_plus": "momentum_buy_signal", "to_minus": "momentum_sell_signal"})
        return df

    def _process(self) -> pd.DataFrame:

        applied_df = self._apply(df=self.series)
        signal_df = self._signal(df=applied_df)
        required_columns = ["momentum", "sma_momentum", "momentum_buy_signal", "momentum_sell_signal"]
        return signal_df[required_columns]



from kabutobashi.domain.entity.blocks.parameterize_blocks import (
    ParameterizeAdxBlock,
    ParameterizeBollingerBandsBlock,
    ParameterizeMacdBlock,
    ParameterizeMomentumBlock,
    ParameterizePsychoLogicalBlock,
    ParameterizeSmaBlock,
    ParameterizeStochasticsBlock,
)
from kabutobashi.domain.entity.blocks.pre_process_blocks import DefaultPreProcessBlock
from kabutobashi.domain.entity.blocks.process_blocks import (
    ProcessAdxBlock,
    ProcessBollingerBandsBlock,
    ProcessMacdBlock,
    ProcessMomentumBlock,
    ProcessPsychoLogicalBlock,
    ProcessSmaBlock,
    ProcessStochasticsBlock,
)
from kabutobashi.domain.entity.blocks.read_blocks import ReadSqlite3Block
from kabutobashi.domain.entity.blocks.reduce_blocks import FullyConnectBlock
from kabutobashi.domain.entity.blocks.write_blocks import WriteImpactSqlite3Block
from kabutobashi.domain.services.flow import Flow


def analysis(code: str, database_dir: str):
    blocks = [
        ReadSqlite3Block,
        DefaultPreProcessBlock,
        ProcessSmaBlock,
        ParameterizeSmaBlock,
        ProcessMacdBlock,
        ParameterizeMacdBlock,
        ProcessAdxBlock,
        ParameterizeAdxBlock,
        ProcessBollingerBandsBlock,
        ParameterizeBollingerBandsBlock,
        ProcessMomentumBlock,
        ParameterizeMomentumBlock,
        ProcessPsychoLogicalBlock,
        ParameterizePsychoLogicalBlock,
        ProcessStochasticsBlock,
        ParameterizeStochasticsBlock,
        FullyConnectBlock,
        WriteImpactSqlite3Block,
    ]

    return Flow.initialize(
        params={
            "read_sqlite3": {"code": code, "database_dir": database_dir},
            "write_impact_sqlite3": {"database_dir": database_dir},
        }
    ).then(blocks)



from .process_adx_block import ProcessAdxBlock
from .process_bollinger_bands_block import ProcessBollingerBandsBlock
from .process_ichimoku import ProcessIchimokuBlock
from .process_macd_block import ProcessMacdBlock
from .process_momentum_block import ProcessMomentumBlock
from .process_psycho_logical_block import ProcessPsychoLogicalBlock
from .process_sma_block import ProcessSmaBlock
from .process_stochastics_block import ProcessStochasticsBlock



from datetime import datetime
from typing import Tuple

import jpholiday
import pandas as pd

from kabutobashi.domain.errors import KabutobashiBlockSeriesDtIsMissingError
from kabutobashi.utilities import convert_float, convert_int, get_working_days_between

from ..decorator import block


@block(
    block_name="default_pre_process",
    series_required_columns=["open", "high", "low", "close", "code", "volume"],
    series_required_columns_mode="all",
)
class DefaultPreProcessBlock:
    series: pd.DataFrame
    for_analysis: bool

    @staticmethod
    def _fix_dt(x: str) -> str:
        try:
            datetime.strptime(x, "%Y-%m-%d")
            return x
        except ValueError:
            pass

        try:
            dt = datetime.strptime(x, "%Y/%m/%d")
            return dt.strftime("%Y-%m-%d")
        except ValueError:
            pass
        return ""

    @staticmethod
    def _is_holiday(x: str) -> bool:
        try:
            dt = datetime.strptime(x, "%Y-%m-%d")
            is_holiday = jpholiday.is_holiday(dt)
            return is_holiday
        except ValueError:
            return False

    @staticmethod
    def _check_dt_requirements(df: pd.DataFrame):
        working_days = get_working_days_between(start_date=min(df["dt"]), end_date=max(df["dt"]))
        working_dt = pd.DataFrame(working_days, columns=["dt"])
        merged = pd.merge(df, working_dt, on="dt", how="outer")
        missing_df = merged[merged["name"].isnull()]
        if not missing_df.empty:
            raise KabutobashiBlockSeriesDtIsMissingError(code=str(list(set(df["code"]))[0]), dt=list(missing_df["dt"]))

    def _process(self) -> Tuple[pd.DataFrame, dict]:

        df = self.series
        # if self.for_analysis:
        #     required_cols = ["open", "high", "low", "close", "code", "volume"]
        #     if df is None:
        #         raise KabutobashiBlockSeriesIsNoneError()
        #     df = df[required_cols]

        columns = df.columns
        df["dt"] = df.index
        df = df.reset_index(drop=True)
        df["dt"] = df["dt"].apply(self._fix_dt)
        df["passing"] = df["dt"].apply(self._is_holiday)
        df["open"] = df["open"].apply(convert_float)
        df["high"] = df["high"].apply(convert_float)
        df["low"] = df["low"].apply(convert_float)
        df["close"] = df["close"].apply(convert_float)
        df["volume"] = df["volume"].apply(convert_int)
        if "pbr" in columns:
            df["pbr"] = df["pbr"].apply(convert_float)
        if "psr" in columns:
            df["psr"] = df["psr"].apply(convert_float)
        if "per" in columns:
            df["per"] = df["per"].apply(convert_float)
        df = df[~df["passing"]]
        # check dt
        self._check_dt_requirements(df=df)
        # remove unused column
        df.index = df["dt"]
        df = df.drop(["passing"], axis=1)
        return df, {"dt": max(df["dt"])}



from .decoded_html_pages import (
    DecodedHtmlPage,
    DecodeHtmlPageStockInfoMinkabuTop,
    DecodeHtmlPageStockInfoMultipleDays,
    DecodeHtmlPageStockIpo,
)
from .raw_html_pages import IHtmlPageRepository, RawHtmlPage, RawHtmlPageStockInfo, RawHtmlPageStockIpo
from .user_agent import UserAgent



from typing import Optional, Tuple

import pandas as pd
from bs4 import BeautifulSoup

from kabutobashi.domain.values import DecodeHtmlPageStockIpo

from ..decorator import block


@block(block_name="extract_stock_ipo", params_required_keys=["html_text"])
class ExtractStockIpoBlock:
    html_text: str

    def _decode(self, html_text: str) -> dict:
        soup = BeautifulSoup(html_text, features="lxml")
        table_content = soup.find("div", {"class": "tablewrap"})
        table_thead = table_content.find("thead")
        # headの取得
        table_head_list = []
        for th in table_thead.find_all("th"):
            table_head_list.append(th.get_text())

        # bodyの取得
        table_tbody = table_content.find("tbody")
        whole_result = []
        for idx, tr in enumerate(table_tbody.find_all("tr")):
            table_body_dict = {}
            for header, td in zip(table_head_list, tr.find_all("td")):
                table_body_dict[header] = td.get_text().replace("\n", "")
            whole_result.append(DecodeHtmlPageStockIpo.from_dict(data=table_body_dict).to_dict())
        return {"ipo_list": whole_result}

    def _process(self) -> Tuple[pd.DataFrame, dict]:
        # to_df
        result = self._decode(html_text=self.html_text)
        df = pd.DataFrame(data=result["ipo_list"])
        return df, result

    def _validate_output(self, series: Optional[pd.DataFrame], params: Optional[dict]):
        keys = series.keys()
        # assert "ipo_list" in keys, "StockIpoExtractBlockOutput must have 'ipo_list' column"



import pandas as pd

from kabutobashi.domain.errors import KabutobashiBlockSeriesIsNoneError

from ..decorator import block
from .abc_parameterize_block import get_impact


@block(
    block_name="parameterize_volatility",
    series_required_columns=["high", "low", "close"],
)
class ParameterizeVolatilityBlock:
    """
    変動幅を計算する
    """

    series: pd.DataFrame

    def _process(self) -> dict:
        df = self.series
        df["volatility"] = (df["high"] - df["low"]) / df["close"]
        volatility_ = df["volatility"].mean()
        close_volatility = max(df["close"]) - min(df["close"]) / df["close"].median()
        return {
            "volatility": volatility_,
            "close_volatility": close_volatility,
            "dt": max(df.index.to_list()),
        }



from dataclasses import dataclass, field, replace
from logging import getLogger
from typing import Dict, List, Literal, Optional, TypeAlias, Union

import pandas as pd

from kabutobashi.domain.errors import KabutobashiBlockGlueError

logger = getLogger(__name__)

__all__ = ["BlockGlue", "BlockOutput", "SeriesRequiredColumnsMode"]
SeriesRequiredColumnsMode: TypeAlias = Literal["strict", "all"]


@dataclass(frozen=True)
class BlockInput:
    series: pd.DataFrame
    params: dict


@dataclass(frozen=True)
class BlockOutput:
    series: Optional[pd.DataFrame]
    params: Optional[dict]
    block_name: str
    execution_order: int = 1


@dataclass(frozen=True)
class IBlock:
    pass


@dataclass(frozen=True)
class SeriesColumns:
    block_name: str
    columns: List[str]
    execution_order: int

    def __lt__(self, other: "SeriesColumns") -> bool:
        if not isinstance(other, SeriesColumns):
            raise ValueError
        return self.execution_order > other.execution_order

    @staticmethod
    def fix(series_columns_list: List["SeriesColumns"]) -> List["SeriesColumns"]:
        res_list = []
        _existed_columns = []
        sorted_list = sorted(series_columns_list)
        for sc in sorted_list:
            unique_elements = list(set(sc.columns).symmetric_difference(set(_existed_columns)) & set(sc.columns))
            _existed_columns.extend(sc.columns)
            res_list.append(replace(sc, columns=unique_elements))
        return res_list


@dataclass(frozen=True)
class BlockGlue:
    series: Optional[pd.DataFrame] = None
    params: Optional[dict] = None
    block_outputs: Dict[str, BlockOutput] = field(default_factory=dict, repr=False)
    execution_order: int = 1

    def get_series_from_required_columns(
        self, required_columns: List[str], series_required_columns_mode: SeriesRequiredColumnsMode = "strict"
    ) -> pd.DataFrame:
        logger.debug(f"{required_columns=}")
        orders = [v.execution_order for _, v in self if v.series is not None]
        if len(orders) != len(set(orders)):
            raise KabutobashiBlockGlueError(f"{orders=} must be unique.")
        series_columns_list = [
            SeriesColumns(block_name=v.block_name, columns=v.series.columns, execution_order=v.execution_order)
            for _, v in self
            if v.series is not None
        ]
        fixed_series_columns_list = SeriesColumns.fix(series_columns_list=series_columns_list)
        initial_series = self[fixed_series_columns_list[0].block_name].series[fixed_series_columns_list[0].columns]
        rest_series = [self[v.block_name].series[v.columns] for v in fixed_series_columns_list[1:]]
        series = initial_series.join(rest_series)
        if series_required_columns_mode == "strict":
            return series[required_columns]
        elif series_required_columns_mode == "all":
            return series
        else:
            raise ValueError()

    def get_series(
        self,
        series_required_columns: Optional[list],
        series_required_columns_mode: SeriesRequiredColumnsMode,
    ) -> Optional[pd.DataFrame]:
        if self.series is not None:
            series = self.series
        elif series_required_columns is not None and type(series_required_columns) is list:
            series = self.get_series_from_required_columns(
                required_columns=series_required_columns, series_required_columns_mode=series_required_columns_mode
            )
        else:
            series = None
        return series

    def get_params(self, block_name: str, params_required_keys: Optional[Union[str, list]]) -> dict:
        params = {}
        if self.params is not None:
            params.update(self.params.get(block_name, {}))
        if params_required_keys is not None and type(params_required_keys) is list:
            logger.debug(f"{params_required_keys=}")
            params = {}
            for k, v in self:
                if v.params is None:
                    continue
                if k == "FLOW_INITIAL":
                    # specific parameters when using Flow.initialize()
                    params.update(v.params.get(block_name, {}))
                else:
                    params.update(v.params)
        return params

    def get_max_execution_order(self) -> int:
        execution_order = [0]
        execution_order.extend([v.execution_order for _, v in self.block_outputs.items()])
        return max(execution_order)

    def __len__(self):
        return len(self.block_outputs.keys())

    def __getitem__(self, key: str):
        if type(key) is str:
            return self.block_outputs[key]
        else:
            raise KeyError(f"Key {key} is not a str")

    def __iter__(self):
        for k, v in self.block_outputs.items():
            yield k, v

    def __contains__(self, item: str):
        if type(item) is str:
            return item in self.block_outputs.keys()
        else:
            raise KeyError(f"Key {item} is not a str")



from typing import Tuple

import pandas as pd

from kabutobashi.infrastructure.repository import KabutobashiDatabase

from ..decorator import block


@block(block_name="read_sqlite3")
class ReadSqlite3Block:
    code: str | int
    database_dir: str

    def _process(self) -> Tuple[pd.DataFrame, dict]:
        df = KabutobashiDatabase(database_dir=self.database_dir).select_stock_df(code=self.code)
        df.index = df["dt"]
        return df, {"code": self.code}

    def _validate_code(self, code: str | int):
        if code is None:
            raise ValueError()



import pandas as pd

from kabutobashi.domain.errors import KabutobashiBlockSeriesIsNoneError

from ..decorator import block


@block(
    block_name="parameterize_pct_change",
    series_required_columns=["close"],
)
class ParameterizePctChangeBlock:
    """
    変化率を計算する
    """

    series: pd.DataFrame

    def _process(self) -> dict:
        df = self.series
        if df is None:
            raise KabutobashiBlockSeriesIsNoneError()
        pct_05 = df["close"].pct_change(5).mean()
        pct_10 = df["close"].pct_change(10).mean()
        pct_20 = df["close"].pct_change(20).mean()
        pct_30 = df["close"].pct_change(30).mean()
        pct_40 = df["close"].pct_change(40).mean()
        return {
            "pct_05": pct_05,
            "pct_10": pct_10,
            "pct_20": pct_20,
            "pct_30": pct_30,
            "pct_40": pct_40,
            "dt": max(df.index.to_list()),
        }



import pandas as pd

from kabutobashi.domain.errors import KabutobashiBlockSeriesIsNoneError

from ..decorator import block
from .abc_parameterize_block import get_impact


@block(
    block_name="parameterize_bollinger_bands",
    series_required_columns=[
        "upper_1_sigma",
        "lower_1_sigma",
        "upper_2_sigma",
        "lower_2_sigma",
        "bollinger_bands_buy_signal",
        "bollinger_bands_sell_signal",
    ],
)
class ParameterizeBollingerBandsBlock:
    series: pd.DataFrame
    influence: int = 2
    tail: int = 5

    def _process(self) -> dict:
        df = self.series
        if df is None:
            raise KabutobashiBlockSeriesIsNoneError()
        params = {
            "upper_1_sigma": df["upper_1_sigma"].tail(3).mean(),
            "lower_1_sigma": df["lower_1_sigma"].tail(3).mean(),
            "upper_2_sigma": df["upper_2_sigma"].tail(3).mean(),
            "lower_2_sigma": df["lower_2_sigma"].tail(3).mean(),
            "bollinger_bands_impact": get_impact(
                df=df, influence=self.influence, tail=self.tail, prefix="bollinger_bands"
            ),
            "dt": max(df.index.to_list()),
        }

        return params



import pandas as pd

from ..decorator import block
from .abc_process_block import cross

__all__ = ["ProcessAdxBlock"]


@block(block_name="process_adx", series_required_columns=["high", "low", "close"])
class ProcessAdxBlock:
    """
    相場のトレンドの強さを見るための指標である`ADX`を計算するクラス。

    以下の指標を計算するクラス

    * +DI: 株価の上昇の大きさ
    * -DI: 株価の下降の大きさ
    * ADX: 株価のトレンドの強さ
    * ADXR: ADXの単純移動平均線

    Args:
        term (int):
        adx_term (int):
        adxr_term (int):

    See Also:
        * https://www.sevendata.co.jp/shihyou/technical/dmi.html
        * https://www.sevendata.co.jp/shihyou/technical/adx.html

    """

    series: pd.DataFrame
    term: int = 14
    adx_term: int = 14
    adxr_term: int = 28

    @staticmethod
    def _true_range(x: pd.DataFrame):
        """

        Args:
            x (pd.DataFrame)

        Returns:
            maximum
        """
        if x["high"] is pd.NA or x["low"] is pd.NA or x["shift_close"] is pd.NA:
            return 0
        current_high = x["high"]
        current_low = x["low"]
        prev_close = x["shift_close"]
        a = current_high - current_low
        b = current_high - prev_close
        c = prev_close - current_low
        max_ab = max(a, b)
        max_ac = max(a, c)
        return max(max_ab, max_ac)

    @staticmethod
    def _compute_dx(x: pd.DataFrame) -> float:
        """

        Args:
            x (pd.DataFrame):
        """
        numerator = abs(x["plus_di"] - x["minus_di"])
        denominator = x["plus_di"] + x["minus_di"]
        try:
            return numerator / denominator * 100
        except ZeroDivisionError:
            return 0

    @staticmethod
    def _fixed_plus_dm(x: pd.DataFrame) -> float:
        if x["plus_dm"] is pd.NA or x["minus_dm"] is pd.NA:
            return 0
        if x["plus_dm"] > 0 and x["plus_dm"] > x["minus_dm"]:
            return x["plus_dm"]
        else:
            return 0

    @staticmethod
    def _fixed_minus_dm(x: pd.DataFrame) -> float:
        if x["plus_dm"] is pd.NA or x["minus_dm"] is pd.NA:
            return 0
        if x["minus_dm"] > 0 and x["minus_dm"] > x["plus_dm"]:
            return x["minus_dm"]
        else:
            return 0

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        # 利用する値をshift
        df = df.assign(shift_high=df["high"].shift(1), shift_low=df["low"].shift(1), shift_close=df["close"].shift(1))
        df = df.assign(
            plus_dm=df.apply(lambda x: x["high"] - x["shift_high"], axis=1),
            minus_dm=df.apply(lambda x: x["shift_low"] - x["low"], axis=1),
        )
        df = df.assign(
            fixed_plus_dm=df.apply(lambda x: self._fixed_plus_dm(x), axis=1),
            fixed_minus_dm=df.apply(lambda x: self._fixed_minus_dm(x), axis=1),
        )
        df = df.assign(
            true_range=df.apply(lambda x: self._true_range(x), axis=1),
            sum_tr=lambda x: x["true_range"].rolling(self.term).sum(),
            sum_plus_dm=lambda x: x["fixed_plus_dm"].rolling(self.term).sum(),
            sum_minus_dm=lambda x: x["fixed_minus_dm"].rolling(self.term).sum(),
        )

        # +DI, -DI
        df = df.assign(
            plus_di=df.apply(lambda x: x["sum_plus_dm"] / x["sum_tr"] * 100, axis=1),
            minus_di=df.apply(lambda x: x["sum_minus_dm"] / x["sum_tr"] * 100, axis=1),
        )
        df = df.assign(
            DX=df.apply(self._compute_dx, axis=1),
            ADX=lambda x: x["DX"].rolling(self.adx_term).mean(),
            ADXR=lambda x: x["DX"].rolling(self.adxr_term).mean(),
        )
        return df

    @staticmethod
    def _buy_signal(x) -> float:
        """
        DMIとADXを組み合わせた基本パターン
        """

        # +DIが-DIを上抜き、ADXが上昇傾向の上向きであれば新規買い
        if x["ADX_trend"] > 0:
            if x["to_plus"] > 0:
                return 1
        return 0

        # +DIが-DIより上に位置している際に、
        # ADXが下向きから上向きに転換した場合
        # if x['diff'] > 0:

    @staticmethod
    def _sell_signal(x) -> float:
        """
        DMIとADXを組み合わせた基本パターン
        """
        # -DIが+DIを下抜き、ADXが下落傾向の下向きであれば新規空売り
        if x["ADX_trend"] < 0:
            if x["to_minus"] > 0:
                return 1
        return 0

    @staticmethod
    def _trend(_s: pd.Series) -> pd.Series:
        """
        ある系列_sのトレンドを計算する。
        差分のrolling_sumを返す
        """
        # shorten variable name
        col = "original"
        shifted = "shifted"

        # shiftしたDataFrameの作成
        shift_s = _s.shift(1)
        df = pd.DataFrame({col: _s, shifted: shift_s})
        df["diff"] = df["original"] - df["shifted"]
        df["diff_rolling_sum"] = df["diff"].rolling(5).sum()
        return df["diff_rolling_sum"]

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        buy_signalとsell_signalを付与
        """
        df["ADX_trend"] = self._trend(df["ADX"])
        df["diff"] = df["plus_di"] - df["minus_di"]
        df = df.join(cross(df["diff"]))

        df["adx_buy_signal"] = df.apply(lambda x: self._buy_signal)
        df["adx_sell_signal"] = df.apply(lambda x: self._sell_signal)

        return df

    def _process(self) -> pd.DataFrame:
        applied_df = self._apply(df=self.series)
        signal_df = self._signal(df=applied_df)
        return signal_df[["DX", "ADX", "ADXR", "adx_buy_signal", "adx_sell_signal"]]



from typing import Tuple

import pandas as pd

from ..decorator import block


@block(
    block_name="fully_connect",
    params_required_keys=[
        "code",
        "dt",
        "sma_impact",
        "macd_impact",
        "adx_impact",
        "bollinger_bands_impact",
        "momentum_impact",
        "psycho_logical_impact",
        "stochastics_impact",
    ],
)
class FullyConnectBlock:

    def _process(self) -> Tuple[pd.DataFrame, dict]:
        impact = (
            self.params["sma_impact"] * self.params.get("sma_impact_ratio", 0.1)
            + self.params["macd_impact"] * self.params.get("macd_impact_ratio", 0.5)
            + self.params["adx_impact"] * self.params.get("adx_impact_ratio", 0.1)
            + self.params["bollinger_bands_impact"] * self.params.get("bollinger_bands_impact_ratio", 0.1)
            + self.params["momentum_impact"] * self.params.get("momentum_impact_ratio", 0.1)
            + self.params["psycho_logical_impact"] * self.params.get("psycho_logical_impact_ratio", 0.1)
            + self.params["stochastics_impact"] * self.params.get("stochastics_impact_ratio", 0.1)
        )
        df = pd.DataFrame.from_dict({"code": [str(self.params["code"])], "dt": [self.params["dt"]], "impact": [impact]})
        return df, {"impact": impact}



from typing import Optional

import pandas as pd

from ..decorator import block
from .abc_crawl_block import from_url


@block(block_name="crawl_stock_info")
class CrawlStockInfoBlock:
    code: str

    def _process(self) -> dict:
        html_text = from_url(url=f"https://minkabu.jp/stock/{self.code}")
        return {"code": self.code, "html_text": html_text}

    def _validate_code(self, code: str):
        if code is None:
            raise ValueError()

    def _validate_output(self, series: Optional[pd.DataFrame], params: Optional[dict]):
        keys = params.keys()
        assert "code" in keys, "CrawlStockInfoBlockOutput must have 'code' column"
        assert "html_text" in keys, "CrawlStockInfoBlockOutput must have 'html_text' column"



import pandas as pd

from ..decorator import block

__all__ = ["ProcessIchimokuBlock"]


@block(block_name="process_ichimoku", series_required_columns=["high", "low", "close"])
class ProcessIchimokuBlock:
    """

    See Also:
        https://kabu.com/investment/guide/technical/04.html
    """

    series: pd.DataFrame
    short_term: int = 12
    medium_term: int = 26
    long_term: int = 52

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(
            # 短期の線
            short_max=lambda x: x["close"].rolling(self.short_term).max(),
            short_min=lambda x: x["close"].rolling(self.short_term).min(),
            # 中期の線
            medium_max=lambda x: x["close"].rolling(self.medium_term).max(),
            medium_min=lambda x: x["close"].rolling(self.medium_term).min(),
            # 長期線
            long_max=lambda x: x["close"].rolling(self.long_term).max(),
            long_min=lambda x: x["close"].rolling(self.long_term).min(),
        )

        # 指標の計算
        df = df.assign(
            line_change=lambda x: (x["short_max"] + x["short_min"]) / 2,
            line_base=lambda x: (x["medium_max"] + x["medium_min"]) / 2,
            # 先行線
            proceeding_span_1=lambda x: (x["line_change"] + x["line_base"]) / 2,
            proceeding_span_2=lambda x: (x["long_max"] + x["long_min"]) / 2,
        )

        # 値のshift
        df = df.assign(
            proceeding_span_1=df["proceeding_span_1"].shift(26),
            proceeding_span_2=df["proceeding_span_2"].shift(26),
            delayed_span=df["close"].shift(26),
        )
        return df

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        return df

    def _process(self) -> pd.DataFrame:
        applied_df = self._apply(df=self.series)
        signal_df = self._signal(df=applied_df)
        return signal_df[["line_change", "line_base", "proceeding_span_1", "proceeding_span_2", "delayed_span"]]



from .application import analysis, crawl_info, crawl_info_multiple, crawl_ipo, crawl_missing_info, decode_brand_list
from .domain import errors
from .domain.entity.blocks import block
from .domain.services.flow import Flow, FlowPath

# methods to analysis
from .domain.values import DecodeHtmlPageStockIpo, RawHtmlPageStockInfo, RawHtmlPageStockIpo
from .example_data import example
from .infrastructure.repository import KabutobashiDatabase

# n営業日前までの日付のリストを返す関数
from .utilities import get_past_n_days

# comparable tuple
VERSION = (0, 8, 6)
# generate __version__ via VERSION tuple
__version__ = ".".join(map(str, VERSION))

# module level doc-string
__doc__ = """
kabutobashi
===========

**kabutobashi** is a Python package to analysis stock data with measure
analysis methods, such as MACD, SMA, etc.

Main Features
-------------
Here are the things that kabutobashi does well:
 - Easy crawl.
 - Easy analysis.
"""



import pandas as pd


def cross(_s: pd.Series, to_plus_name=None, to_minus_name=None) -> pd.DataFrame:
    """
    0を基準としてプラスかマイナスのどちらかに振れたかを判断する関数

    Args:
        _s: 対象のpd.Series
        to_plus_name: 上抜けた場合のカラムの名前
        to_minus_name: 下抜けた場合のカラムの名前
    """
    # shorten variable name
    col = "original"
    shifted = "shifted"

    # shiftしたDataFrameの作成
    shift_s = _s.shift(1)
    df = pd.DataFrame({col: _s, shifted: shift_s})

    # 正負が交差した点
    df = df.assign(
        is_cross=df.apply(lambda x: 1 if x[col] * x[shifted] < 0 else 0, axis=1),
        is_higher=df.apply(lambda x: 1 if x[col] > x[shifted] else 0, axis=1),
        is_lower=df.apply(lambda x: 1 if x[col] < x[shifted] else 0, axis=1),
    )

    # 上抜けか下抜けかを判断している
    df = df.assign(to_plus=df["is_cross"] * df["is_higher"], to_minus=df["is_cross"] * df["is_lower"])
    if to_plus_name is not None:
        df = df.rename(columns={"to_plus": to_plus_name})
    if to_minus_name is not None:
        df = df.rename(columns={"to_minus": to_minus_name})
    return df



import pandas as pd

from kabutobashi.infrastructure.repository import KabutobashiDatabase

from ..decorator import block


@block(
    block_name="write_stock_sqlite3",
    series_required_columns=["code", "dt", "name", "open", "close", "high", "low", "volume"],
    series_required_columns_mode="all",
    params_required_keys=["database_dir"],
)
class WriteStockSqlite3Block:
    series: pd.DataFrame
    database_dir: str

    def _process(self) -> dict:
        KabutobashiDatabase(database_dir=self.database_dir).insert_stock_df(df=self.series)
        return {"status": "success"}


@block(
    block_name="write_impact_sqlite3",
    series_required_columns=["code", "dt", "impact"],
    series_required_columns_mode="strict",
    params_required_keys=["database_dir"],
)
class WriteImpactSqlite3Block:
    series: pd.DataFrame
    database_dir: str

    def _process(self) -> dict:
        KabutobashiDatabase(database_dir=self.database_dir).insert_impact_df(df=self.series)
        return {"status": "success"}


@block(
    block_name="write_brand_sqlite3",
    series_required_columns=["code", "name", "market", "industry_type"],
    series_required_columns_mode="strict",
    params_required_keys=["database_dir"],
)
class WriteBrandSqlite3Block:
    series: pd.DataFrame
    database_dir: str

    def _process(self) -> dict:
        db = KabutobashiDatabase(database_dir=self.database_dir)
        current_brand_df = db.select_brand_df()
        current_code_list = current_brand_df["code"].to_list()
        insert_candidate_series = self.series[~self.series["code"].isin(current_code_list)]
        if not insert_candidate_series.empty:
            db.insert_brand_df(df=insert_candidate_series)
        return {"status": "success"}



from .extract_stock_info_block import ExtractStockInfoBlock
from .extract_stock_info_multiple_days_block import (
    ExtractStockInfoMultipleDays2Block,
    ExtractStockInfoMultipleDaysBlock,
)
from .extract_stock_ipo_block import ExtractStockIpoBlock



import pandas as pd

from ..decorator import block
from .abc_process_block import cross

__all__ = ["ProcessSmaBlock"]


@block(block_name="process_sma", series_required_columns=["close"])
class ProcessSmaBlock:
    series: pd.DataFrame
    short_term: int = 5
    medium_term: int = 21
    long_term: int = 70

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:

        df = df.assign(
            sma_short=df["close"].rolling(self.short_term).mean(),
            sma_medium=df["close"].rolling(self.medium_term).mean(),
            sma_long=df["close"].rolling(self.long_term).mean(),
        )
        return df

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        df["diff"] = df.apply(lambda x: x["sma_long"] - x["sma_short"], axis=1)
        # 正負が交差した点
        df = df.join(cross(df["diff"]))
        df = df.rename(columns={"to_plus": "sma_buy_signal", "to_minus": "sma_sell_signal"})
        return df

    def _process(self) -> pd.DataFrame:
        applied_df = self._apply(df=self.series)
        signal_df = self._signal(df=applied_df)
        return signal_df[["sma_short", "sma_medium", "sma_long", "sma_buy_signal", "sma_sell_signal"]]



import re
from dataclasses import dataclass
from datetime import datetime
from logging import getLogger
from typing import Dict, List, Union

from kabutobashi.domain.values import (
    DecodeHtmlPageStockInfoMinkabuTop,
    DecodeHtmlPageStockIpo,
    RawHtmlPageStockInfo,
    RawHtmlPageStockIpo,
)

from .utils import IHtmlDecoder, PageDecoder

logger = getLogger(__name__)


@dataclass(frozen=True)
class StockInfoMinkabuTopHtmlDecoder(IHtmlDecoder):
    """
    Model: Service(Implemented)

    Examples:
        >>> from kabutobashi.infrastructure.repository import StockInfoHtmlPageRepository
        >>> # get single page
        >>> html_page = StockInfoHtmlPageRepository(code="0001").read()
        >>> result = StockInfoMinkabuTopHtmlDecoder().decode_to_dict(html_page=html_page)
    """

    def _decode_to_object_hook(self, data: dict) -> DecodeHtmlPageStockInfoMinkabuTop:
        return DecodeHtmlPageStockInfoMinkabuTop.from_dict(data=data)

    def _decode(self, html_page: RawHtmlPageStockInfo) -> dict:
        soup = html_page.get_as_soup()
        result: Dict[str, Union[str, bool, int, float, List[str]]] = {"html": html_page.html}

        stock_board_tag = "md_stockBoard"

        raw_dt = PageDecoder(tag1="span", class1="fsm").decode(bs=soup)
        pattern = r"\((?P<month>[0-9]+)/(?P<day>[0-9]+)\)|\((?P<hour>[0-9]+):(?P<minute>[0-9]+)\)"
        match_result = re.match(pattern, raw_dt)
        dt = datetime.now()
        if match_result:
            rep = match_result.groupdict()
            if rep.get("month"):
                dt = dt.replace(month=int(rep["month"]))
            if rep.get("day"):
                dt = dt.replace(day=int(rep["day"]))

        # ページ上部の情報を取得
        stock_board = soup.find("div", {"class": stock_board_tag})
        result.update(
            {
                "stock_label": PageDecoder(tag1="div", class1="stock_label").decode(bs=stock_board),
                "name": PageDecoder(tag1="p", class1="md_stockBoard_stockName").decode(bs=stock_board),
                "close": PageDecoder(tag1="div", class1="stock_price").decode(bs=stock_board),
            }
        )

        # ページ中央の情報を取得
        stock_detail = soup.find("div", {"id": "main"})
        info = {}
        for li in stock_detail.find_all("tr", {"class": "ly_vamd"}):
            info[li.find("th").get_text()] = li.find("td").get_text()
        stock_label = str(result.get("stock_label", ""))
        code, market = stock_label.split("  ")
        result.update(
            {
                "dt": dt.strftime("%Y-%m-%d"),
                "code": code,
                "industry_type": PageDecoder(tag1="div", class1="ly_content_wrapper size_ss").decode(bs=stock_detail),
                "market": market,
                "open": info.get("始値", "0"),
                "high": info.get("高値", "0"),
                "low": info.get("安値", "0"),
                "unit": info.get("単元株数", "0"),
                "per": info.get("PER(調整後)", "0"),
                "psr": info.get("PSR", "0"),
                "pbr": info.get("PBR", "0"),
                "volume": info.get("出来高", "0"),
                "market_capitalization": info.get("時価総額", "---"),
                "issued_shares": info.get("発行済株数", "---"),
            }
        )

        return result


@dataclass(frozen=True)
class StockIpoHtmlDecoder(IHtmlDecoder):
    """
    Model: Service(Implemented)
    """

    def _decode(self, html_page: RawHtmlPageStockIpo) -> dict:
        soup = html_page.get_as_soup()
        table_content = soup.find("div", {"class": "tablewrap"})
        table_thead = table_content.find("thead")
        # headの取得
        table_head_list = []
        for th in table_thead.find_all("th"):
            table_head_list.append(th.get_text())

        # bodyの取得
        table_tbody = table_content.find("tbody")
        whole_result = []
        for idx, tr in enumerate(table_tbody.find_all("tr")):
            table_body_dict = {}
            for header, td in zip(table_head_list, tr.find_all("td")):
                table_body_dict[header] = td.get_text().replace("\n", "")
            whole_result.append(DecodeHtmlPageStockIpo.from_dict(data=table_body_dict).to_dict())
        return {"ipo_list": whole_result}

    def _decode_to_object_hook(self, data: dict) -> List[DecodeHtmlPageStockIpo]:
        ipo_list = data["ipo_list"]
        result_list = []
        for v in ipo_list:
            result_list.append(DecodeHtmlPageStockIpo.from_dict(data=v))
        return result_list



import math

import pandas as pd

from ..decorator import block


@block(block_name="process_stochastics", series_required_columns=["close", "low", "high"])
class ProcessStochasticsBlock:
    series: pd.DataFrame

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        df_ = df.copy()
        df_["K"] = self._fast_stochastic_k(df_["close"], df_["low"], df_["high"], 9)
        df_["D"] = self._fast_stochastic_d(df_["K"])
        df_["SD"] = self._slow_stochastic_d(df_["D"])
        return df_

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        買いと売りに関する指標を算出する
        """
        df = df.assign(
            shift_K=lambda x: x["K"].shift(1),
            shift_D=lambda x: x["D"].shift(1),
            shift_SD=lambda x: x["SD"].shift(1),
        ).fillna(0)

        # 複数引数は関数を利用することで吸収
        df["stochastics_buy_signal"] = df.apply(self._buy_signal_index_internal, axis=1)
        df["stochastics_sell_signal"] = df.apply(self._sell_signal_index_internal, axis=1)
        return df

    @staticmethod
    def _fast_stochastic_k(close, low, high, n):
        return (
            (close - low.rolling(window=n, center=False).min())
            / (high.rolling(window=n, center=False).max() - low.rolling(window=n, center=False).min())
        ) * 100

    @staticmethod
    def _fast_stochastic_d(stochastic_k):
        # ストキャスの%Dを計算（%Kの3日SMA）
        return stochastic_k.rolling(window=3, center=False).mean()

    @staticmethod
    def _slow_stochastic_d(stochastic_d):
        # ストキャスの%SDを計算（%Dの3日SMA）
        return stochastic_d.rolling(window=3, center=False).mean()

    @staticmethod
    def _buy_signal_index_internal(x: pd.Series) -> float:
        return ProcessStochasticsBlock._buy_signal_index(
            x["K"], x["D"], x["SD"], x["shift_K"], x["shift_D"], x["shift_SD"]
        )

    @staticmethod
    def _buy_signal_index(current_k, current_d, current_sd, prev_k, prev_d, prev_sd) -> float:
        if (current_k > 30) | (current_d > 30) | (current_sd > 30):
            return 0

        # %K・%D共に20％以下の時に、%Kが%Dを下から上抜いた時
        if current_k < 20 and current_d < 20:
            if (prev_d > prev_k) and (current_d < current_k):
                return current_k - current_d

        # %D・スロー%D共に20％以下の時に、%Dがスロー%Dを下から上抜いた時
        if current_d < 20 and current_sd < 20:
            if (prev_sd > prev_d) and (current_sd < current_d):
                return current_d - current_sd
        return 1 / math.exp(
            math.pow(current_k - 20, 2) / 100 + math.pow(current_d - 20, 2) / 100 + math.pow(current_sd - 20, 2) / 100
        )

    @staticmethod
    def _sell_signal_index_internal(x: pd.Series) -> float:
        return ProcessStochasticsBlock._sell_signal_index(
            x["K"], x["D"], x["SD"], x["shift_K"], x["shift_D"], x["shift_SD"]
        )

    @staticmethod
    def _sell_signal_index(current_k, current_d, current_sd, prev_k, prev_d, prev_sd) -> float:
        if (current_k < 70) | (current_d < 70) | (current_sd < 70):
            return 0

        # %K・%D共に80％以上の時に、%Kが%Dを上から下抜いた時
        if current_k > 80 and current_d > 80:
            if (prev_d < prev_k) and (current_d > current_k):
                return current_d - current_k

        # %D・スロー%D共に80％以上の時に、%Dがスロー%Dを上から下抜いた時
        # %D・スロー%D共に20％以下の時に、%Dがスロー%Dを下から上抜いた時
        if current_d > 80 and current_sd > 80:
            if (prev_sd < prev_d) and (current_sd > current_d):
                return current_d - current_sd
        return 1 / math.exp(
            math.pow(current_k - 20, 2) / 100 + math.pow(current_d - 20, 2) / 100 + math.pow(current_sd - 20, 2) / 100
        )

    def _process(self) -> pd.DataFrame:
        applied_df = self._apply(df=self.series)
        signal_df = self._signal(df=applied_df)
        required_columns = ["K", "D", "SD", "stochastics_buy_signal", "stochastics_sell_signal"]
        return signal_df[required_columns]



from typing import Optional

import pandas as pd

from ..decorator import block
from .abc_crawl_block import from_url


@block(block_name="crawl_stock_info_multiple_days")
class CrawlStockInfoMultipleDaysBlock:
    code: str

    def _process(self) -> dict:
        main_html_text = from_url(url=f"https://minkabu.jp/stock/{self.code}/daily_bar")
        sub_html_text = from_url(url=f"https://minkabu.jp/stock/{self.code}/daily_valuation")
        return {"code": self.code, "main_html_text": main_html_text, "sub_html_text": sub_html_text}

    def _validate_output(self, series: Optional[pd.DataFrame], params: Optional[dict]):
        keys = params.keys()
        assert "code" in keys, "StockInfoMultipleDaysCrawlBlockOutput must have 'code' column"
        assert "main_html_text" in keys, "StockInfoMultipleDaysCrawlBlockOutput must have 'main_html_text' column"
        assert "sub_html_text" in keys, "StockInfoMultipleDaysCrawlBlockOutput must have 'sub_html_text' column"


@block(block_name="crawl_stock_info_multiple_days_2")
class CrawlStockInfoMultipleDays2Block:
    code: str
    page: int

    def _process(self) -> dict:
        main_text = from_url(url=f"https://kabutan.jp/stock/kabuka?code={self.code}&ashi=day&page={self.page}")
        return {"code": self.code, "main_html_text": main_text}

    def _validate_output(self, series: Optional[pd.DataFrame], params: Optional[dict]):
        keys = params.keys()
        assert "code" in keys, "StockInfoMultipleDaysCrawlBlockOutput must have 'code' column"
        assert "main_html_text" in keys, "StockInfoMultipleDaysCrawlBlockOutput must have 'main_html_text' column"



import pandas as pd

from kabutobashi.infrastructure.repository import KabutobashiDatabase


def decode_brand_list(path: str, database_dir: str) -> pd.DataFrame:
    """
    See Also: https://www.jpx.co.jp/markets/statistics-equities/misc/01.html
    """
    df = pd.read_excel(path)
    column_renames = {
        "コード": "code",
        "銘柄名": "name",
        "市場・商品区分": "market",
        "33業種区分": "industry_type",
    }
    df = df.rename(columns=column_renames)
    df = df[column_renames.values()]
    df["market"] = df["market"].apply(lambda x: x.replace("（内国株式）", ""))
    prime_df = df[df["market"] == "プライム"]
    standard_df = df[df["market"] == "スタンダード"]
    growth_df = df[df["market"] == "グロース"]
    merged_df = pd.concat([prime_df, standard_df, growth_df]).reset_index()

    KabutobashiDatabase(database_dir=database_dir).insert_brand_df(df=merged_df)
    return merged_df



import re
from datetime import datetime
from typing import Dict, List, Optional, Union

import pandas as pd
from bs4 import BeautifulSoup

from kabutobashi.domain.services.decode_html.utils import PageDecoder

from ..decorator import block


@block(block_name="extract_stock_info", params_required_keys=["html_text"])
class ExtractStockInfoBlock:
    html_text: str

    def _decode(self, html_text: str) -> dict:
        soup = BeautifulSoup(html_text, features="lxml")
        result: Dict[str, Union[str, bool, int, float, List[str]]] = {}

        stock_board_tag = "md_stockBoard"

        raw_dt = PageDecoder(tag1="span", class1="fsm").decode(bs=soup)
        if isinstance(raw_dt, list):
            raise ValueError()
        pattern = r"\((?P<month>[0-9]+)/(?P<day>[0-9]+)\)|\((?P<hour>[0-9]+):(?P<minute>[0-9]+)\)"
        match_result = re.match(pattern=pattern, string=raw_dt)
        dt = datetime.now()
        if match_result:
            rep = match_result.groupdict()
            if rep.get("month"):
                dt = dt.replace(month=int(rep["month"]))
            if rep.get("day"):
                dt = dt.replace(day=int(rep["day"]))

        # ページ上部の情報を取得
        stock_board = soup.find("div", {"class": stock_board_tag})
        result.update(
            {
                "stock_label": PageDecoder(tag1="div", class1="stock_label").decode(bs=stock_board),
                "name": PageDecoder(tag1="p", class1="md_stockBoard_stockName").decode(bs=stock_board),
                "close": PageDecoder(tag1="div", class1="stock_price").decode(bs=stock_board),
            }
        )

        # ページ中央の情報を取得
        stock_detail = soup.find("div", {"id": "main"})
        info = {}
        for li in stock_detail.find_all("tr", {"class": "ly_vamd"}):
            info[li.find("th").get_text()] = li.find("td").get_text()
        stock_label = str(result.get("stock_label", ""))
        code, market = stock_label.split("  ")
        result.update(
            {
                "dt": dt.strftime("%Y-%m-%d"),
                "code": code,
                "industry_type": PageDecoder(tag1="div", class1="ly_content_wrapper size_ss").decode(bs=stock_detail),
                "market": market,
                "open": info.get("始値", "0"),
                "high": info.get("高値", "0"),
                "low": info.get("安値", "0"),
                "unit": info.get("単元株数", "0"),
                "per": info.get("PER(調整後)", "0"),
                "psr": info.get("PSR", "0"),
                "pbr": info.get("PBR", "0"),
                "volume": info.get("出来高", "0"),
                "market_capitalization": info.get("時価総額", "---"),
                "issued_shares": info.get("発行済株数", "---"),
            }
        )

        return result

    def _process(self) -> pd.DataFrame:
        result = self._decode(html_text=self.html_text)
        # to_df
        df = pd.DataFrame(data=result, index=[result["dt"]])
        return df

    def _validate_output(self, series: Optional[pd.DataFrame], params: Optional[dict]):
        keys = series.keys()
        assert "code" in keys, "StockInfoExtractBlockOutput must have 'code' column"
        assert "stock_label" in keys, "StockInfoExtractBlockOutput must have 'stock_label' column"
        assert "name" in keys, "StockInfoExtractBlockOutput must have 'name' column"
        assert "dt" in keys, "StockInfoExtractBlockOutput must have 'dt' column"
        assert "industry_type" in keys, "StockInfoExtractBlockOutput must have 'industry_type' column"
        assert "close" in keys, "StockInfoExtractBlockOutput must have 'close' column"
        assert "market" in keys, "StockInfoExtractBlockOutput must have 'market' column"
        assert "open" in keys, "StockInfoExtractBlockOutput must have 'open' column"
        assert "high" in keys, "StockInfoExtractBlockOutput must have 'high' column"
        assert "low" in keys, "StockInfoExtractBlockOutput must have 'low' column"
        assert "unit" in keys, "StockInfoExtractBlockOutput must have 'unit' column"
        assert "per" in keys, "StockInfoExtractBlockOutput must have 'per' column"
        assert "psr" in keys, "StockInfoExtractBlockOutput must have 'psr' column"
        assert "pbr" in keys, "StockInfoExtractBlockOutput must have 'pbr' column"
        assert "volume" in keys, "StockInfoExtractBlockOutput must have 'volume' column"
        assert "market_capitalization" in keys, "StockInfoExtractBlockOutput must have 'market_capitalization' column"
        assert "issued_shares" in keys, "StockInfoExtractBlockOutput must have 'issued_shares' column"



import pandas as pd

from kabutobashi.domain.errors import KabutobashiBlockSeriesIsNoneError

from ..decorator import block
from .abc_parameterize_block import get_impact


@block(
    block_name="parameterize_psycho_logical",
    series_required_columns=["psycho_line", "psycho_logical_buy_signal", "psycho_logical_sell_signal"],
)
class ParameterizePsychoLogicalBlock:
    series: pd.DataFrame
    influence: int = 2
    tail: int = 5

    def _process(self) -> dict:
        df = self.series
        if df is None:
            raise KabutobashiBlockSeriesIsNoneError()
        params = {
            "psycho_line": df["psycho_line"].tail(3).mean(),
            "psycho_logical_impact": get_impact(
                df=df, influence=self.influence, tail=self.tail, prefix="psycho_logical"
            ),
            "dt": max(df.index.to_list()),
        }

        return params



import os
from pathlib import Path
from typing import Tuple

import pandas as pd

from ..decorator import block

PARENT_PATH = os.path.abspath(os.path.dirname(__file__))
PACKAGE_ROOT = Path(PARENT_PATH).parent.parent.parent.parent.parent
DATA_PATH = f"{PACKAGE_ROOT}/data"


@block(block_name="read_example")
class ReadExampleBlock:
    code: str | int

    def _process(self) -> Tuple[pd.DataFrame, dict]:
        file_name = "example.csv.gz"
        df = pd.read_csv(f"{DATA_PATH}/{file_name}")
        df = df[df["code"] == self.code]
        df.index = df["dt"]
        return df, {"code": self.code}

    def _validate_code(self, code: str | int):
        if code is None:
            raise ValueError()



import re
import warnings
from dataclasses import dataclass
from functools import partial
from inspect import signature
from logging import getLogger
from types import FunctionType
from typing import Iterator, List, NoReturn, Optional, Tuple, TypeAlias, Union

import pandas as pd

from kabutobashi.domain.errors import (
    KabutobashiBlockDecoratorNameError,
    KabutobashiBlockDecoratorNotImplementedError,
    KabutobashiBlockDecoratorReturnError,
    KabutobashiBlockDecoratorTypeError,
)

from .basis_blocks import BlockGlue, BlockOutput, SeriesRequiredColumnsMode

__all__ = ["block", "block_from"]

blocks_dict = {}
logger = getLogger(__name__)

# type candidates of `UdfBlock._process()` return
BlockProcessResultType: TypeAlias = Union[dict, pd.DataFrame, Tuple[dict, pd.DataFrame], Tuple[pd.DataFrame, dict]]


@dataclass(frozen=True)
class SeriesRequiredColumn:
    block_name: str
    keys: List[str]
    priority: int


@dataclass(frozen=True)
class ParamsRequiredKey:
    block_name: str
    keys: List[str]
    priority: int


def _to_snake_case(string: str) -> str:
    # see: https://qiita.com/munepi0713/items/82ce7a56aa1b8233fd30
    _PARSE_BY_SEP_PATTERN = re.compile(r"[ _-]+")
    _PARSE_PATTERN = re.compile(r"[A-Za-z][^A-Z]+")

    def _parse_words(_string: str) -> Iterator[str]:
        for block in re.split(_PARSE_BY_SEP_PATTERN, _string):
            for m in re.finditer(_PARSE_PATTERN, block):
                yield m.group(0)

    word_iter = _parse_words(string)
    return "_".join(word.lower() for word in word_iter)


def _set_qualname(cls, value):
    # Ensure that the functions returned from _create_fn uses the proper
    # __qualname__ (the class they belong to).
    if isinstance(value, FunctionType):
        value.__qualname__ = f"{cls.__qualname__}.{value.__name__}"
    return value


def _set_new_attribute(cls, name, value):
    # Never overwrites an existing attribute.  Returns True if the
    # attribute already exists.
    if name in cls.__dict__:
        return True
    _set_qualname(cls, value)
    setattr(cls, name, value)
    return False


def _inner_func_validate_input(self, *, functions: List[callable] = None) -> NoReturn:
    """
    The method is NOT intended to override by users.
    """
    logger.debug(f"{self.params=}")
    for f in functions:
        sig = signature(f)
        function_required_args = sig.parameters.keys()
        function_args = {k: v for k, v in self.__dict__.items() if (k in function_required_args)}
        logger.debug(f"{f} <= {function_args=}")
        f(self, **function_args)


def _inner_func_validate_output(self, series: Optional[pd.DataFrame], params: Optional[dict]) -> NoReturn:
    """
    The method is NOT intended to override by users.
    """
    self._validate_output(series=series, params=params)


def _inner_func_private_validate_output(self, series: Optional[pd.DataFrame], params: Optional[dict]) -> NoReturn:
    """
    Default _validate_output() method.
    The method is intended to override by users.
    """
    pass


def _inner_func_process(self) -> BlockGlue:
    """
    The method is NOT intended to override by users.
    The method can be called by `cls.process()`

    Returns:
        BlockGlue
    """
    # initialize
    block_name = self.block_name
    res_glue = BlockGlue()
    if self._glue:
        res_glue = self._glue
    # validate_input
    self.validate_input()
    # process()
    res: BlockProcessResultType = self._process()
    execution_order = res_glue.get_max_execution_order() + 1

    if type(res) is tuple:
        if len(res) == 2:
            if type(res[0]) is dict and type(res[1]) is pd.DataFrame:
                self.validate_output(series=res[1], params=res[0])
                block_output = BlockOutput(
                    series=res[1], params=res[0], block_name=block_name, execution_order=execution_order
                )
            elif type(res[1]) is dict and type(res[0]) is pd.DataFrame:
                self.validate_output(series=res[0], params=res[1])
                block_output = BlockOutput(
                    series=res[0], params=res[1], block_name=block_name, execution_order=execution_order
                )
            else:
                raise KabutobashiBlockDecoratorReturnError(
                    "The return values are limited to combinations of `dict` and `pd.DataFrame`."
                )
        else:
            raise KabutobashiBlockDecoratorReturnError("Please limit the number of return values to two or fewer.")
    elif type(res) is dict:
        self.validate_output(series=None, params=res)
        block_output = BlockOutput(series=None, params=res, block_name=block_name, execution_order=execution_order)
    elif type(res) is pd.DataFrame:
        self.validate_output(series=res, params=None)
        block_output = BlockOutput(series=res, params=None, block_name=block_name, execution_order=execution_order)
    else:
        raise KabutobashiBlockDecoratorReturnError(f"An unexpected return type {type(res)} was returned.")
    block_outputs = res_glue.block_outputs if res_glue.block_outputs else {}
    block_outputs.update({block_name: block_output})
    return BlockGlue(
        series=res_glue.series,
        params=res_glue.params,
        block_outputs=block_outputs,
        execution_order=res_glue.execution_order + 1,
    )


def _inner_class_func_factory(cls, glue: BlockGlue):
    """
    The method is NOT intended to override by users.
    The method can be called by `cls.factory()`

    Returns:
        cls()
    """
    setattr(cls, "_glue", glue)
    # get parameters from glue
    params = {}
    series = None

    res = cls._factory(glue)
    if type(res) is tuple:
        if len(res) == 2:
            if type(res[0]) is dict:
                params, series = res
            elif type(res[1]) is dict:
                series, params = res
            else:
                raise KabutobashiBlockDecoratorReturnError(
                    "The return values are limited to combinations of `dict` and `pd.DataFrame`."
                )
        else:
            raise KabutobashiBlockDecoratorReturnError("Please limit the number of return values to two or fewer.")
    elif type(res) is dict:
        params = res
    elif type(res) is pd.DataFrame:
        series = res
    else:
        raise KabutobashiBlockDecoratorReturnError(f"An unexpected return type {type(res)} was returned.")

    # set attributes
    logger.debug(f"@block._factory(): {cls.__name__}: {params.keys()}")
    for k, v in params.items():
        setattr(cls, k, v)
    return cls(series=series, params=params)


def _inner_class_default_private_func_factory(cls, glue: BlockGlue) -> Tuple[pd.DataFrame, dict]:
    """
    Default _factory() method.
    Although the method is intended to override by users, usually the method is not overridden.
    The method can be called by `cls._factory()`

    Returns:
        cls()
    """
    cls_instance = cls()
    block_name = cls_instance.block_name
    series_required_columns = cls_instance.series_required_columns
    params_required_keys = cls_instance.params_required_keys
    series_required_columns_mode = cls_instance.series_required_columns_mode

    # params
    params = glue.get_params(
        block_name=block_name,
        params_required_keys=params_required_keys,
    )

    # series
    series = glue.get_series(
        series_required_columns=series_required_columns,
        series_required_columns_mode=series_required_columns_mode,
    )
    return series, params


def _inner_class_func_glue(cls, glue: BlockGlue) -> BlockGlue:
    """
    The method is NOT intended to override by users.

    Returns:
        BlockGlue()
    """
    block_instance = cls.factory(glue=glue)
    new_glue = block_instance.process()
    return new_glue


def _inner_init(self, series: Optional[pd.DataFrame] = None, params: Optional[dict] = None):
    """
    The method is NOT intended to override by users.
    """
    self.series = series
    self.params = params


def _inner_repr(self):
    """
    The method is NOT intended to override by users.
    """
    # block name
    repr = ["# block_name: {self.block_name}"]
    # attributes
    repr.extend([f"+ {name}: {value}" for name, value in self.__dict__.items()])
    # repr.extend([f"+ {name}: {getattr(self, name)}" for name in self.__annotations__.keys()])

    return "\n".join(repr)


def _process_class(
    cls,
    block_name: str,
    factory: bool,
    process: bool,
    series_required_columns: List[str | SeriesRequiredColumn],
    series_required_columns_mode: SeriesRequiredColumnsMode,
    params_required_keys: List[str | ParamsRequiredKey],
):
    cls_params = {}
    cls_annotations = cls.__dict__.get("__annotations__", {})
    logger.debug(f"{cls_annotations=}")
    if not cls.__name__.endswith("Block"):
        raise KabutobashiBlockDecoratorNameError(f"class name must end with 'Block', {cls.__name__} is not allowed.")

    cls_keys = cls.__dict__.keys()
    # check _process
    if process:
        if "_process" not in cls_keys:
            raise KabutobashiBlockDecoratorNotImplementedError("_process method is not implemented.")
        if not isinstance((cls.__dict__["_process"]), FunctionType):
            raise ValueError()
        _process_annotation_candidates = [Tuple[dict, pd.DataFrame], Tuple[pd.DataFrame, dict], dict, pd.DataFrame]
        # check annotation types
        _process_annotations = cls.__dict__["_process"].__annotations__
        if "return" in _process_annotations:
            if not any([_process_annotations["return"] is t for t in _process_annotation_candidates]):
                warn_msg = f"{_process_annotations['return']} is not compatible. Use `dict`, `pd.DataFrame`, or `Tuple[dict, pd.DataFrame]`"
                warnings.warn(warn_msg, category=SyntaxWarning)

    # check _factory
    if factory:
        if "_factory" not in cls_keys:
            raise KabutobashiBlockDecoratorNotImplementedError("_factory method is not implemented.")
        if not type(cls.__dict__["_factory"]) is classmethod:
            raise ValueError()

    for name in cls_annotations:
        if name in cls.__dict__.keys():
            cls_params[name] = cls.__dict__[name]
        else:
            cls_params[name] = None

    # set-block-name
    if block_name is None:
        block_name = _to_snake_case(cls.__name__)
    setattr(cls, "block_name", block_name)
    setattr(cls, "series_required_columns", series_required_columns)
    setattr(cls, "params_required_keys", params_required_keys)
    setattr(cls, "series_required_columns_mode", series_required_columns_mode)
    # set-params
    setattr(cls, "params", cls_params)
    # process function
    _set_new_attribute(cls=cls, name="process", value=_inner_func_process)
    # factory function
    _set_new_attribute(cls=cls, name="factory", value=classmethod(_inner_class_func_factory))
    if not factory:
        _set_new_attribute(cls=cls, name="_factory", value=classmethod(_inner_class_default_private_func_factory))
    # operate function
    _set_new_attribute(cls=cls, name="glue", value=classmethod(_inner_class_func_glue))
    # validation functions
    _inner_func_validate_input_partial = partial(
        _inner_func_validate_input,
        self=cls,
        functions=[
            v for k, v in cls.__dict__.items() if k.startswith("_validate") and not k.startswith("_validate_output")
        ],
    )
    _set_new_attribute(cls=cls, name="validate_input", value=_inner_func_validate_input_partial)
    _set_new_attribute(cls=cls, name="validate_output", value=_inner_func_validate_output)
    if "_validate_output" not in cls.__dict__:
        _set_new_attribute(cls=cls, name="_validate_output", value=_inner_func_private_validate_output)
    # dunder-method
    _set_new_attribute(cls=cls, name="__init__", value=_inner_init)
    _set_new_attribute(cls=cls, name="__repr__", value=_inner_repr)
    # register global dict
    blocks_dict.update({block_name: cls})
    return cls


def block(
    cls=None,
    /,
    *,
    block_name: str = None,
    factory: bool = False,
    process: bool = True,
    series_required_columns: List[str | SeriesRequiredColumn] = None,
    series_required_columns_mode: SeriesRequiredColumnsMode = "strict",
    params_required_keys: List[str | ParamsRequiredKey] = None,
):
    """

    Args:
        cls: class to decorate
        block_name: BlockName
        factory: True if _factory() method is required to implement.
        process: True if _process() method is required to implement.
        series_required_columns:
        series_required_columns_mode:
        params_required_keys:

    Returns:
        decorator

    Examples:
        >>> # basic example
        >>> from kabutobashi.domain.entity.blocks import BlockGlue
        >>> @block()
        >>> class SampleBlock:
        >>>     term: int = 10
        >>>
        >>>     @classmethod
        >>>     def _factory(cls, glue: BlockGlue) -> "SampleBlock":
        >>>         return SampleBlock()
        >>>
        >>>     def _process(self):
        >>>         params = self
        >>>         return SampleBlock()
    """

    def wrap(_cls):
        if type(_cls) is not type:
            raise KabutobashiBlockDecoratorTypeError(f"first argument of @block must be a class, not a {type(_cls)}")
        return _process_class(
            _cls,
            block_name=block_name,
            factory=factory,
            process=process,
            series_required_columns=series_required_columns,
            series_required_columns_mode=series_required_columns_mode,
            params_required_keys=params_required_keys,
        )

    # See if we're being called as @dataclass or @dataclass().
    if cls is None:
        # We're called with parens.
        return wrap
    return wrap(cls)


def block_from(key: str):
    return blocks_dict[key]



import pandas as pd

from kabutobashi.domain.errors import KabutobashiBlockSeriesIsNoneError

from ..decorator import block
from .abc_parameterize_block import get_impact


@block(
    block_name="parameterize_adx", series_required_columns=["DX", "ADX", "ADXR", "adx_buy_signal", "adx_sell_signal"]
)
class ParameterizeAdxBlock:
    series: pd.DataFrame
    influence: int = 2
    tail: int = 5

    def _process(self) -> dict:
        df = self.series
        if df is None:
            raise KabutobashiBlockSeriesIsNoneError()
        params = {
            "adx_dx": df["DX"].tail(3).mean(),
            "adx_adx": df["ADX"].tail(3).mean(),
            "adx_adxr": df["ADXR"].tail(3).mean(),
            "adx_impact": get_impact(df=df, influence=self.influence, tail=self.tail, prefix="adx"),
            "dt": max(df.index.to_list()),
        }
        return params



from typing import Optional

import pandas as pd

from ..decorator import block
from .abc_crawl_block import from_url


@block(block_name="crawl_stock_ipo")
class CrawlStockIpoBlock:
    year: str

    def _process(self) -> dict:
        html_text = from_url(url=f"https://96ut.com/ipo/list.php?year={self.year}")
        return {"year": self.year, "html_text": html_text}

    def _validate_output(self, series: Optional[pd.DataFrame], params: Optional[dict]):
        keys = params.keys()
        assert "year" in keys, "StockIpoCrawlBlockOutput must have 'year' column"
        assert "html_text" in keys, "StockIpoCrawlBlockOutput must have 'html_text' column"



import pandas as pd

from kabutobashi.domain.errors import KabutobashiBlockSeriesIsNoneError

from ..decorator import block
from .abc_parameterize_block import get_impact


@block(
    block_name="parameterize_stochastics",
    series_required_columns=["K", "D", "SD", "stochastics_buy_signal", "stochastics_sell_signal"],
)
class ParameterizeStochasticsBlock:
    series: pd.DataFrame
    influence: int = 2
    tail: int = 5

    def _process(self) -> dict:
        df = self.series
        if df is None:
            raise KabutobashiBlockSeriesIsNoneError()
        params = {
            "stochastics_k": df["K"].tail(3).mean(),
            "stochastics_d": df["D"].tail(3).mean(),
            "stochastics_sd": df["SD"].tail(3).mean(),
            "stochastics_impact": get_impact(df=df, influence=self.influence, tail=self.tail, prefix="stochastics"),
            "dt": max(df.index.to_list()),
        }

        return params



import pandas as pd


def get_impact(df: pd.DataFrame, influence: int, tail: int, prefix: str = "") -> float:
    """
    売りと買いのシグナルの余波の合計値を返す。

    Args:
        df:
        influence:
        tail:

    Returns:
        [-1,1]の値をとる。-1: 売り、1: 買いを表す
    """
    columns = df.columns
    if "buy_signal" not in columns and f"{prefix}_buy_signal" not in columns:
        return 0
    if "sell_signal" not in columns and f"{prefix}_sell_signal" not in columns:
        return 0

    if "buy_signal" in columns:
        df["buy_impact"] = df["buy_signal"].ewm(span=influence).mean()
    elif f"{prefix}_buy_signal" in columns:
        df["buy_impact"] = df[f"{prefix}_buy_signal"].ewm(span=influence).mean()

    if "sell_signal" in columns:
        df["sell_impact"] = df["sell_signal"].ewm(span=influence).mean()
    elif f"{prefix}_sell_signal" in columns:
        df["sell_impact"] = df[f"{prefix}_sell_signal"].ewm(span=influence).mean()

    buy_impact_index = df["buy_impact"].iloc[-tail:].sum()
    sell_impact_index = df["sell_impact"].iloc[-tail:].sum()
    return round(buy_impact_index - sell_impact_index, 5)



from .parameterize_adx_block import ParameterizeAdxBlock
from .parameterize_bollinger_bands_block import ParameterizeBollingerBandsBlock
from .parameterize_macd_block import ParameterizeMacdBlock
from .parameterize_momentum_block import ParameterizeMomentumBlock
from .parameterize_pct_change import ParameterizePctChangeBlock
from .parameterize_psycho_logical_block import ParameterizePsychoLogicalBlock
from .parameterize_sma_block import ParameterizeSmaBlock
from .parameterize_stochastics_block import ParameterizeStochasticsBlock
from .parameterize_volatility import ParameterizeVolatilityBlock



import pandas as pd

from ..decorator import block
from .abc_process_block import cross

__all__ = ["ProcessMacdBlock"]


@block(block_name="process_macd", series_required_columns=["close"])
class ProcessMacdBlock:
    series: pd.DataFrame
    short_term: int = 12
    long_term: int = 26
    macd_span: int = 9

    def _apply(self, df: pd.DataFrame) -> pd.DataFrame:
        df = df.assign(
            # MACDの計算
            ema_short=lambda x: x["close"].ewm(span=self.short_term).mean(),
            ema_long=lambda x: x["close"].ewm(span=self.long_term).mean(),
            macd=lambda x: x["ema_short"] - x["ema_long"],
            signal=lambda x: x["macd"].ewm(span=self.macd_span).mean(),
            # ヒストグラム値
            histogram=lambda x: x["macd"] - x["signal"],
        )
        return df

    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        # 正負が交差した点
        df = df.join(cross(df["histogram"]))
        df = df.rename(columns={"to_plus": "macd_buy_signal", "to_minus": "macd_sell_signal"})
        return df

    def _process(self) -> pd.DataFrame:
        applied_df = self._apply(df=self.series)
        signal_df = self._signal(df=applied_df)
        return signal_df[
            ["ema_short", "ema_long", "macd", "signal", "histogram", "macd_buy_signal", "macd_sell_signal"]
        ]



import sqlite3
from logging import INFO, getLogger
from pathlib import Path
from typing import Optional

import pandas as pd

logger = getLogger(__name__)
logger.setLevel(INFO)

ROOT_PATH = Path(__file__).parent.parent.parent.parent


class KabutobashiDatabase:
    def __init__(self, database_dir: str = ROOT_PATH, database_name: str = "kabutobashi.db"):
        self.database_dir = database_dir
        self.database_name = database_name
        self.con = None

    def __enter__(self):
        self.con = sqlite3.connect(f"{self.database_dir}/{self.database_name}")
        return self.con

    def __exit__(self, ex_type, ex_value, trace):
        self.con.commit()
        self.con.close()

    def initialize(self) -> "KabutobashiDatabase":
        # stock daily record
        drop_stock_table = "DROP TABLE IF EXISTS stock"
        create_stock_statement = """
            CREATE TABLE IF NOT EXISTS stock(
                code TEXT NOT NULL,
                dt TEXT NOT NULL,
                name TEXT,
                open REAL,
                high REAL,
                low REAL,
                close REAL,
                volume REAL,
                PRIMARY KEY (code, dt)
            )
            """
        create_stock_index_statement = "CREATE INDEX IF NOT EXISTS stock_code_dt_idx ON stock (code, dt)"

        # evaluate
        drop_impact_table = "DROP TABLE IF EXISTS impact"
        create_impact_statement = """
            CREATE TABLE IF NOT EXISTS impact(
                code TEXT NOT NULL,
                dt TEXT NOT NULL,
                impact REAL,
                PRIMARY KEY (code, dt)
            )
            """
        create_impact_index_statement = "CREATE INDEX IF NOT EXISTS impact_code_dt_idx ON impact (code, dt)"

        # brand
        drop_brand_table = "DROP TABLE IF EXISTS brand"
        create_brand_statement = """
            CREATE TABLE IF NOT EXISTS brand(
                code TEXT NOT NULL,
                name TEXT,
                market TEXT,
                industry_type TEXT,
                PRIMARY KEY (code)
            )
            """
        create_brand_index_statement = "CREATE INDEX IF NOT EXISTS brand_code_dt_idx ON brand (code)"
        with self as conn:
            cur = conn.cursor()
            cur.execute(drop_stock_table)
            cur.execute(create_stock_statement)
            cur.execute(create_stock_index_statement)
            cur.execute(drop_impact_table)
            cur.execute(create_impact_statement)
            cur.execute(create_impact_index_statement)
            cur.execute(drop_brand_table)
            cur.execute(create_brand_statement)
            cur.execute(create_brand_index_statement)
        return self

    def insert_stock_df(self, df: pd.DataFrame) -> "KabutobashiDatabase":
        stock_table_columns = ["code", "dt", "name", "open", "close", "high", "low", "volume"]
        stock_table_name = "stock"
        with self as conn:
            df = df.reset_index(drop=True)
            try:
                df[stock_table_columns].to_sql(stock_table_name, conn, if_exists="append", index=False)
            except sqlite3.IntegrityError:
                logger.warning(f"stock_df(stock.code, stock.dt) already exists, {df=}")
        return self

    def select_stock_df(self, code: str):
        stock_table_columns = ["code", "dt", "name", "open", "close", "high", "low", "volume"]
        with self as conn:
            df = pd.read_sql(f"SELECT * FROM stock WHERE code = '{code}'", conn)
            return df[stock_table_columns]

    def insert_impact_df(self, df: pd.DataFrame) -> "KabutobashiDatabase":
        impact_table_columns = ["code", "dt", "impact"]
        stock_table_name = "impact"
        with self as conn:
            df = df.reset_index(drop=True)
            try:
                df[impact_table_columns].to_sql(stock_table_name, conn, if_exists="append", index=False)
            except sqlite3.IntegrityError:
                logger.warning(f"impact_df(stock.code, stock.dt) already exists")
        return self

    def select_impact_df(self, dt: str) -> Optional[pd.DataFrame]:
        impact_table_columns = ["code", "dt", "impact"]
        with self as conn:
            try:
                df = pd.read_sql(f"SELECT * FROM impact WHERE dt = '{dt}' ORDER BY impact", conn)
                df["dt"] = df["dt"].astype(str)
                return df[impact_table_columns]
            except sqlite3.DatabaseError:
                return None

    def insert_brand_df(self, df: pd.DataFrame) -> "KabutobashiDatabase":
        impact_table_columns = ["code", "name", "market", "industry_type"]
        stock_table_name = "brand"
        with self as conn:
            df = df.reset_index(drop=True)
            try:
                df[impact_table_columns].to_sql(stock_table_name, conn, if_exists="append", index=False)
            except sqlite3.IntegrityError:
                logger.warning(f"brand_df(brand.code) already exists")
        return self

    def select_brand_df(self) -> Optional[pd.DataFrame]:
        impact_table_columns = ["code", "name", "market", "industry_type"]
        with self as conn:
            try:
                df = pd.read_sql(f"SELECT * FROM brand ORDER BY code", conn)
                return df[impact_table_columns]
            except sqlite3.DatabaseError:
                return None



import pandas as pd

from kabutobashi.domain.entity.blocks.crawl_blocks import *
from kabutobashi.domain.entity.blocks.extract_blocks import *
from kabutobashi.domain.entity.blocks.pre_process_blocks import *
from kabutobashi.domain.entity.blocks.read_blocks import ReadSqlite3Block
from kabutobashi.domain.entity.blocks.write_blocks import *
from kabutobashi.domain.errors import KabutobashiBlockSeriesDtIsMissingError
from kabutobashi.domain.services.flow import Flow
from kabutobashi.infrastructure.repository import KabutobashiDatabase


def crawl_info(code: str, database_dir: str):
    blocks = [
        CrawlStockInfoBlock,
        ExtractStockInfoBlock,
        DefaultPreProcessBlock,
        WriteStockSqlite3Block,
    ]

    res = Flow.initialize(
        params={
            "crawl_stock_info": {"code": code},
            "default_pre_process": {"for_analysis": False},
            "write_stock_sqlite3": {"database_dir": database_dir},
        }
    ).then(blocks)
    return res.block_glue["default_pre_process"].series


def crawl_info_multiple(code: str, page: str, database_dir: str) -> pd.DataFrame:
    blocks = [
        CrawlStockInfoMultipleDays2Block,
        ExtractStockInfoMultipleDays2Block,
        DefaultPreProcessBlock,
        WriteStockSqlite3Block,
    ]

    res = Flow.initialize(
        params={
            "crawl_stock_info_multiple_days_2": {"code": code, "page": page},
            "default_pre_process": {"for_analysis": False},
            "write_stock_sqlite3": {"database_dir": database_dir},
        }
    ).then(blocks)
    return res.block_glue["default_pre_process"].series


def crawl_ipo(year: str, database_dir: str):
    blocks = [
        CrawlStockIpoBlock,
        ExtractStockIpoBlock,
        WriteBrandSqlite3Block,
    ]

    res = Flow.initialize(
        params={
            "crawl_stock_ipo": {"year": year},
            "write_brand_sqlite3": {"database_dir": database_dir},
        }
    ).then(blocks)
    return res.block_glue["extract_stock_ipo"].series


def crawl_missing_info(code: str, database_dir: str):
    detect_blocks = [
        ReadSqlite3Block,
        DefaultPreProcessBlock,
    ]

    try:
        Flow.initialize(
            params={
                "read_sqlite3": {"code": code, "database_dir": database_dir},
                "write_impact_sqlite3": {"database_dir": database_dir},
            }
        ).then(detect_blocks)
    except KabutobashiBlockSeriesDtIsMissingError as e:

        blocks = [
            CrawlStockInfoMultipleDays2Block,
            ExtractStockInfoMultipleDays2Block,
            DefaultPreProcessBlock,
        ]

        res = Flow.initialize(
            params={
                "crawl_stock_info_multiple_days_2": {"code": code, "page": "1"},
                "default_pre_process": {"for_analysis": False},
            }
        ).then(blocks)
        df = res.block_glue["default_pre_process"].series
        df = df.reset_index(drop=True)
        missing_dt = pd.DataFrame(e.dt, columns=["dt"])
        merged = pd.merge(df, missing_dt, on="dt", how="inner")

        db = KabutobashiDatabase(database_dir=database_dir)
        db.insert_stock_df(df=merged)
        return {"status": "success", "rows": len(merged.index)}
    return {"status": "success", "rows": 0}
